import gp
import warnings


class SessionList:
    """
    Keeps a list of all currently registered GenePattern server sessions
    """
    sessions = []

    def register(self, server, username, password):
        """
        Register a new GenePattern server session for the provided
        server, username and password. Return the session.
        :param server:
        :param username:
        :param password:
        :return:
        """

        # Create the session
        session = gp.GPServer(server, username, password)

        # Validate username if not empty
        valid_username = username != "" and username is not None

        # Validate that the server is not already registered
        index = self._get_index(server)
        new_server = index == -1

        # Add the new session to the list
        if valid_username and new_server:
            self.sessions.append(session)

        # Replace old session is one exists
        if valid_username and not new_server:
            self.sessions[index] = session

        return session

    def get(self, server):
        """
        Returns a registered GPServer object with a matching GenePattern server url or index
        Returns None if no matching result was found
        :param server:
        :return:
        """

        # Handle indexes
        if isinstance(server, int):
            if server >= len(self.sessions):
                return None
            else:
                return self.sessions[server]

        # Handle server URLs
        index = self._get_index(server)
        if index == -1:
            return None
        else:
            return self.sessions[index]

    def make(self, server):
        """
        Returns the registered session, if one exists.
        Otherwise, returns a placeholder session with the URL but no credentials
        :param server:
        :return:
        """
        session = self.get(server)
        if session: return session
        elif isinstance(server, int): raise RuntimeError('make() does not support session indexes')
        else: return gp.GPServer(server, None, None)

    def clean(self):
        """
        Clear all GenePattern sessions from the sessions list
        :return:
        """
        self.sessions = []

    def _get_index(self, server_url):
        """
        Returns a registered GPServer object with a matching GenePattern server url
        Returns -1 if no matching result was found
        :param server_url:
        :return:
        """
        for i in range(len(self.sessions)):
            session = self.sessions[i]
            if session.url == server_url:
                return i
        return -1


"""
GenePattern Sessions Singleton
"""
session = SessionList()


def get_session(index):
    """Deprecated: genepattern.get_session shim added for the purposes of backward compatibility"""
    warnings.warn("genepattern.get_session() has been deprecated. Please use genepattern.session.get() instead.", UserWarning)
    return session.get(index)


def register_session(server, username, password):
    """Deprecated call for backward compatibility"""
    warnings.warn("register_session will be deprecated, use sessions.register instead", UserWarning)
    return session.register(server, username, password)



from .authwidget import GENEPATTERN_SERVERS, GPAuthWidget
from .taskwidget import GPTaskWidget, reproduce_job, load_task
from .jobwidget import GPJobWidget
from .sessions import session, get_session, register_session
from .display import display
from nbtools import UIBuilder as GPUIBuilder, UIOutput as GPUIOutput, build_ui, open

__author__ = 'Thorin Tabor'
__copyright__ = 'Copyright 2014-2023, Regents of the University of California & Broad Institute'
__version__ = '23.11'
__status__ = 'Production'
__license__ = 'BSD-3-Clause'



import inspect
import json
import os
from urllib.request import Request, urlopen
from urllib.error import HTTPError
from gp import GPTask
from IPython.display import display
from ipywidgets import Output
from .jobwidget import GPJobWidget
from nbtools import NBTool, UIBuilder, UIOutput, python_safe, EventManager
from .shim import get_task, get_kinds, get_eula, accept_eula, job_params, param_groups, job_group
from .utils import GENEPATTERN_LOGO, session_color, server_name


class GPTaskWidget(UIBuilder):
    """A widget for representing the status of a GenePattern job"""
    session_color = None
    task = None
    function_wrapper = None
    parameter_spec = None
    upload_callback = None
    kwargs = {}

    def create_function_wrapper(self, task):
        """Create a function that accepts the expected input and submits a GenePattern job"""
        if task is None or task.server_data is None: return lambda: None  # Dummy function for null task
        name_map = {}  # Map of Python-safe parameter names to GP parameter names

        # Function for submitting a new GenePattern job based on the task form
        def submit_job(**kwargs):
            spec = task.make_job_spec()
            for name, value in kwargs.items():
                if value is None: value = ''    # Handle the case of blank optional parameters
                spec.set_parameter(name_map[name], value)
            job = task.server_data.run_job(spec, wait_until_done=False)
            display(GPJobWidget(job, logo='none', color=session_color(self.task.server_data.url, secondary_color=True)))

        # Function for adding a parameter with a safe name
        def add_param(param_list, p):
            safe_name = python_safe(p.name)
            name_map[safe_name] = p.name
            param = inspect.Parameter(safe_name, inspect.Parameter.POSITIONAL_OR_KEYWORD)
            param_list.append(param)

        # Generate function signature programmatically
        submit_job.__qualname__ = task.name
        submit_job.__doc__ = task.description
        params = []
        params_for_job = task.job_params if hasattr(task, 'job_params') else job_params(task)
        for p in task.params + params_for_job: add_param(params, p)  # Loop over all parameters
        submit_job.__signature__ = inspect.Signature(params)

        return submit_job

    def add_type_spec(self, task_param, param_spec):
        if task_param.attributes['type'] == 'java.io.File':
            param_spec['type'] = 'file'
            if task_param.is_choice_param():
                param_spec['choices'] = {c['label']: c['value'] for c in task_param.get_choices()}
            if task_param.allow_multiple(): param_spec['maximum'] = 100
        elif task_param.is_choice_param():
            param_spec['type'] = 'choice'
            param_spec['choices'] = {c['label']: c['value'] for c in task_param.get_choices()}
            if task_param.allow_choice_custom_value(): param_spec['combo'] = True
            if task_param.allow_multiple(): param_spec['multiple'] = True
        elif task_param.attributes['type'] == 'java.lang.Integer': param_spec['type'] = 'number'
        elif task_param.attributes['type'] == 'java.lang.Float': param_spec['type'] = 'number'
        elif task_param.attributes['type'].lower() == 'password': param_spec['type'] = 'password'
        else: param_spec['type'] = 'text'

    @staticmethod
    def override_if_set(safe_name, attr, param_overrides, param_val):
        if param_overrides and safe_name in param_overrides and attr in param_overrides[safe_name]:
            return param_overrides[safe_name][attr]
        else: return param_val

    def create_param_spec(self, task, kwargs):
        """Create the display spec for each parameter"""
        if task is None: return {}  # Dummy function for null task
        spec = {}
        params_for_job = task.job_params if hasattr(task, 'job_params') else job_params(task)
        param_overrides = kwargs.pop('parameters', None)
        for p in task.params + params_for_job:
            safe_name = python_safe(p.name)
            spec[safe_name] = {}
            spec[safe_name]['default'] = GPTaskWidget.form_value(
                GPTaskWidget.override_if_set(safe_name, 'default', param_overrides, p.get_default_value())
            )
            spec[safe_name]['description'] = GPTaskWidget.form_value(
                GPTaskWidget.override_if_set(safe_name, 'description', param_overrides, p.description)
            )
            spec[safe_name]['optional'] = GPTaskWidget.override_if_set(safe_name, 'optional', param_overrides,
                                                                       p.is_optional())
            spec[safe_name]['kinds'] = GPTaskWidget.override_if_set(safe_name, 'kinds', param_overrides,
                                                                    p.get_kinds() if hasattr(p, 'get_kinds') else get_kinds(p))
            self.add_type_spec(p, spec[safe_name])
        return spec

    @staticmethod
    def extract_parameter_groups(task):
        groups = task.param_groups if hasattr(task, 'param_groups') else param_groups(task)     # Get param groups
        job_options_group = task.job_group if hasattr(task, 'job_group') else job_group(task)   # Get job options
        job_options_group['advanced'] = True                                                    # Hide by default
        all_groups = groups + [job_options_group]                                               # Join groups
        for group in all_groups:                                                                # Escape param names
            if 'parameters' in group:
                for i in range(len(group['parameters'])):
                    group['parameters'][i] = python_safe(group['parameters'][i])
        return all_groups

    def generate_upload_callback(self):
        """Create an upload callback to pass to file input widgets"""
        def genepattern_upload_callback(values):
            try:
                for k in values:
                    path = os.path.realpath(k)
                    gpfile = self.task.server_data.upload_file(k, path)
                    os.remove(path)
                    return gpfile.get_url()
            except Exception as e:
                self.error = f"Error encountered uploading file: {e}"
        return genepattern_upload_callback

    def handle_error_task(self, error_message, name='GenePattern Module', **kwargs):
        """Display an error message if the task is None"""
        ui_args = {'color': session_color(), **kwargs}
        UIBuilder.__init__(self, lambda: None, **ui_args)

        self.name = name
        self.display_header = False
        self.display_footer = False
        self.error = error_message

    def __init__(self, task=None, origin='', id='', **kwargs):
        """Initialize the task widget"""
        self.task = task
        self.kwargs = kwargs
        if task and origin is None: origin = task.server_data.url
        if task and id is None: id = task.lsid

        if self.task is None or self.task.server_data is None:  # Set the right look and error message if task is None
            self.handle_error_task('No GenePattern module specified.', **kwargs)
        elif self.is_java_visualizer():  # Checks if deprecated visualizer and displays an error message
            self.handle_error_task('Java-based visualizers are deprecated in GenePattern and will not function in Jupyter', name=task.name, **kwargs)
        else:
            if self.task.params is None: self.task.param_load()                 # Load params from GP server
            self.function_wrapper = self.create_function_wrapper(self.task)     # Create run task function
            self.parameter_spec = self.create_param_spec(self.task, kwargs)     # Create the parameter spec
            self.session_color = session_color(self.task.server_data.url)       # Set the session color
            ui_args = {                                                         # Assemble keyword arguments
                'color': self.session_color,
                'id': id,
                'license': self.add_license(),
                'license_callback': self.generate_license_callback(),
                'logo': GENEPATTERN_LOGO,
                'origin': origin,
                'parameter_groups': GPTaskWidget.extract_parameter_groups(self.task),
                'parameters': self.parameter_spec,
                'subtitle': f'Version {task.version}',
                'upload_callback': self.generate_upload_callback(),
            }
            ui_args = { **ui_args, **kwargs }                                   # Merge kwargs (allows overrides)
            UIBuilder.__init__(self, self.function_wrapper, **ui_args)          # Initiate the widget
            self.attach_menu_items()

        # Register the event handler for GP login
        EventManager.instance().register("gp.login", self.login_callback)

    def add_license(self):
        if not self.task or not self.task.json: return {}   # Skip if there is no task json to parse
        eula = self.task.get_eula() if hasattr(self.task, 'get_eula') else get_eula(self.task)
        if len(eula['pendingEulas']) == 0: return {}        # No licenses to add

        return {
            'name': f'You must agree to the following end-user license agreement before you can run {self.task.name}.',
            'text': eula['pendingEulas'][0]['content'],
            'callback': False
        }

    def generate_license_callback(self):
        """Function to call when a license is agreed to"""
        return (lambda: self.task.accept_eula()) if hasattr(self.task, 'accept_eula') else (lambda: accept_eula(self.task))

    def attach_menu_items(self):
        """Attach the menu items needed for GenePattern tasks"""
        self.extra_menu_items = {
            'Documentation': {          # Add the documentation link
                'action': 'javascript',
                'code': f'window.open("{self._generate_doc_link()}")'
            }
        }

    def _generate_doc_link(self):
        """Return the documentation URL if absolute; if relative, attach to server URL"""
        if self.task.documentation and self.task.documentation.startswith('/'):
            return f'{self.task.server_data.url[:-3]}{self.task.documentation}'
        else:
            return {self.task.documentation}

    @staticmethod
    def form_value(raw_value):
        """Give the default parameter value in format the UI Builder expects"""
        if raw_value is not None: return raw_value
        else: return ''

    def is_java_visualizer(self):
        if self.task.params is None: self.task.param_load()  # Load params from GP server
        return 'Visualizer' in self.task.dto['categories']

    def login_callback(self, data):
        """Callback upon authentication for unauthenticated task widgets.
           Normally this is handled in the `nbtools.register` callback, but for older notebooks with programmatically
               defined GPTaskWidget cells, this additional callback is necessary to ensure that they render
               appropriately upon login. This is why we check for self.error, as we can safely skip all other cells."""
        if self.task is not None and self.task.server_data is None and self.error is not None:
            task = data.get_task(self.task.uri) if hasattr(data, 'get_task') else get_task(data, self.task.uri)
            task.param_load()  # Get the GPTask object and load its data

            # Create a new task widget and close the old unauthenticated one
            with self.output:
                display(GPTaskWidget(task, **self.kwargs))
                self.close()


class TaskTool(NBTool):
    """Tool wrapper for the authentication widget"""

    def __init__(self, server_name, task):
        NBTool.__init__(self)
        self.origin = server_name
        self.id = task.lsid
        self.name = task.name
        self.description = task.description
        self.load = lambda **kwargs: GPTaskWidget(task, id=self.id, origin=self.origin, **kwargs)


def reproduce_job(sessions, session_index, job_number):
    session = sessions.make(session_index)
    origin = server_name(session.url)
    needs_rendered = True

    # Create the task widget, either immediately or upon login callback
    def create_task():
        session = sessions.make(session_index)          # Retrieve session again, necessary if run from login callback
        request = Request(session.url + "/rest/v1/jobs/" + str(job_number) + "?includeInputParams=true")
        if session.authorization_header() is not None:  # Make authorized call to GenePattern REST API
            request.add_header('Authorization', session.authorization_header())
        request.add_header('User-Agent', 'GenePatternRest')
        response = urlopen(request)                     # Get the response
        json_str = response.read().decode('utf-8')      # Deal with the JSON payload
        json_dict = json.loads(json_str)
        task = GPTask(session, json_dict['taskLsid'])
        task.param_load()                               # Retrieve task parameters
        params = task.get_parameters()
        for i in range(len(params)):
            task_param = params[i]
            custom_value = json_dict['inputParams'][i]  # Set the values to reproduce the original job
            task_param.attributes['default_value'] = custom_value[list(custom_value)[0]]
        return GPTaskWidget(task)                       # Return widget for display

    # Render the widget upon callback from 'nbtools.register'
    def render_task(data):
        nonlocal needs_rendered
        if needs_rendered and 'origin' in data and data['origin'] == origin:
            task_widget = create_task()             # Create the task widget
            placeholder.close()                     # Remove the placeholder widget
            with output: display(task_widget)       # Display the task widget
            needs_rendered = False                  # Only display once

    if not session.username:                        # If not logged in, display placeholder widget
        output = Output()  # Output widget and placeholder
        placeholder = UIOutput(name='Cannot find module', error=f'Cannot find module: {origin} | {job_number}')
        output.append_display_data(placeholder)
        EventManager.instance().register("nbtools.register", render_task)  # Register the callback with event manager
        return output                               # Return widget for display
    else:
        return create_task()


def spec_to_kwargs(kwargs):
    """If a GPJobSpec is included in the kwargs, merge it into the parameters kwarg"""
    if 'spec' in kwargs:
        base_parameters = kwargs['parameters'] if 'parameters' in kwargs else {}
        for p in kwargs['spec'].params:
            if p['name'] not in base_parameters:
                base_parameters[python_safe(p['name'])] = {
                    'default': p['values'] if len(p['values']) > 1 else p['values'][0]
                }
        kwargs['parameters'] = base_parameters  # Update parameters kwarg
        kwargs.pop('spec')                      # Remove spec kwarg


def load_task(sessions, session_index, name_or_lsid):
    """Return a task widget for the specified module name or lsid,
       regardless of whether it's been registered with the ToolManager or not.
       Useful for loading old versions of modules."""

    session = sessions.make(session_index)          # Get the GenePattern session
    task = GPTask(session, name_or_lsid)            # Initialize a task object

    try: task.param_load()                          # Query the GenePattern server for the module's metadata
    except HTTPError: return None                   # Return None if no module available

    return GPTaskWidget(task)                       # Return the task widget



from io import StringIO
from html.parser import HTMLParser
from gp import GPTask, GPTaskParam
import json
import requests
import urllib.parse
import urllib.request


def get_token(session):
    if not hasattr(session, 'token') or not session.token:
        session.token = session.login()
    return session.token


def get_kinds(param):
    if 'fileFormat' not in param.attributes:
        return []  # Default to an empty list

    # Get the list from the attributes
    kinds_string = param.attributes['fileFormat']

    if kinds_string == '' or kinds_string == '*':
        return []  # Accepts all known kinds

    return kinds_string.split(';')


def get_permissions(job):
    url = f'{job.server_data.url}/rest/v1/jobs/{job.job_number}/permissions'
    request = urllib.request.Request(url)
    if job.server_data.authorization_header() is not None:
        request.add_header('Authorization', job.server_data.authorization_header())
    request.add_header('User-Agent', 'GenePatternRest')

    response = urllib.request.urlopen(request)
    return json.loads(response.read())


def set_permissions(job, permissions):
    url = f'{job.server_data.url}/rest/v1/jobs/{job.job_number}/permissions'
    data = json.dumps(permissions).encode('utf8')
    request = urllib.request.Request(url, data=data, method='PUT')
    if job.server_data.authorization_header() is not None:
        request.add_header('Authorization', job.server_data.authorization_header())
    request.add_header('User-Agent', 'GenePatternRest')
    urllib.request.urlopen(request)


def login(session):
    safe_url = ensure_safe_url(session.url)
    safe_username = urllib.parse.quote(session.username)
    safe_password = urllib.parse.quote(session.password)

    url = f"{safe_url}/rest/v1/oauth2/token?grant_type=password&username={safe_username}&password={safe_password}&client_id=GenePatternNotebook-{safe_username}"
    response = requests.post(url)

    try:
        response.raise_for_status()
        return response.json()['access_token']
    except requests.exceptions.Timeout:
        raise TimeoutError('Connection timed out attempting to contact GenePattern server')
    except requests.exceptions.TooManyRedirects:
        raise TimeoutError('Bad GenePattern server URL')
    except requests.exceptions.RequestException as e:
        raise TimeoutError('Invalid username or password')


def system_message(session):
    safe_url = ensure_safe_url(session.url)
    url = f"{safe_url}/rest/v1/config/system-message"
    response = requests.get(url)
    return strip_html(response.text)


def get_task(session, lsid_or_name):
    url = f'{session.url}/rest/v1/tasks/{lsid_or_name}/'
    request = urllib.request.Request(url)
    if session.authorization_header() is not None:
        request.add_header('Authorization', session.authorization_header())
    request.add_header('User-Agent', 'GenePatternRest')

    response = urllib.request.urlopen(request)
    task_dict = json.loads(response.read())
    return GPTask(session, task_dict['lsid'], task_dict)


class HTMLStripper(HTMLParser):
    """Parse HTML blob and strip out all tags"""
    def __init__(self):
        super().__init__()
        self.reset()
        self.strict = False
        self.convert_charrefs = True
        self.text = StringIO()

    def error(self, message):
        pass

    def handle_data(self, d):
        self.text.write(d)

    def get_data(self):
        return self.text.getvalue()


def strip_html(html):
    """Strip potentially unsafe HTML from the system message string"""
    s = HTMLStripper()
    s.feed(html)
    return s.get_data()


def ensure_safe_url(url):
    """Ensure the GenePattern URL ends with /gp"""
    if url.endswith('/'):
        url = url[:-1]
    if not url.endswith('/gp'):
        url += '/gp'
    return url


def get_eula(task):
    """Return a dict containing the module's EULA information"""
    return task.dto['eulaInfo']


def accept_eula(task):
    """Call the EULA accept endpoint"""
    eula = get_eula(task)
    requests.put(eula['acceptUrl'], data=eula['acceptData'],
                 auth=(task.server_data.username, task.server_data.password))
    task.param_load()


def job_params(task):
    if 'config' in task.dto and 'job.inputParams' in task.dto['config']:
        return [GPTaskParam(task, p) for p in task.dto['config']['job.inputParams']]
    else:
        return []


def param_groups(task):
    if 'paramGroups' in task.dto: return task.dto['paramGroups']
    else: return []


def job_group(task):
    if 'config' in task.dto and 'job.inputParamGroup' in task.dto['config']:
        return task.dto['config']['job.inputParamGroup']
    else:
        return {}


def terminate_job(job):
    url = f'{job.server_data.url}/rest/v1/jobs/{job.job_number}/terminate'
    request = urllib.request.Request(url, method='DELETE')
    if job.server_data.authorization_header() is not None:
        request.add_header('Authorization', job.server_data.authorization_header())
    request.add_header('User-Agent', 'GenePatternRest')
    return urllib.request.urlopen(request).code == 200



from threading import Timer
from urllib.error import HTTPError
from urllib.parse import urlparse, parse_qs, urlencode
from ipywidgets import Dropdown, Button, HTML, VBox, HBox
from nbtools import UIOutput, EventManager, ToolManager
from .shim import get_permissions, set_permissions, get_token, terminate_job
from .utils import GENEPATTERN_LOGO, server_name, session_color, is_url, redirect_url


class GPJobWidget(UIOutput):
    """A widget for representing the status of a GenePattern job"""
    sharing_displayed = False
    job = None

    def __init__(self, job=None, **kwargs):
        """Initialize the job widget"""
        self.job = job
        self.set_color(kwargs)
        self.set_logo(kwargs)
        UIOutput.__init__(self, origin=self.job_origin(), **kwargs)
        self.poll()  # Query the GP server and begin polling, if needed
        self.attach_detach()
        self.attach_sharing()
        self.attach_terminate()
        self.attach_duplicate()

        # Register the event handler for GP login
        EventManager.instance().register("gp.login", self.login_callback)

    def job_origin(self):
        if self.job and self.job.server_data: return server_name(self.job.server_data.url)
        else: return ''

    def set_color(self, kwargs={}):
        # If color is already set, keep that color
        if 'color' in kwargs: return

        # Otherwise, set the color based on the session
        if self.job and self.job.server_data: kwargs['color'] = session_color(self.job.server_data.url)
        else: kwargs['color'] = session_color()

        return kwargs['color']

    @staticmethod
    def set_logo(kwargs={}):
        # If logo is already set, keep that logo
        if 'logo' in kwargs: return

        # Otherwise, set the logo
        kwargs['logo'] = GENEPATTERN_LOGO
        return kwargs['logo']

    def poll(self):
        """Poll the GenePattern server for the job info and display it in the widget"""
        if self.initialized():
            try:  # Attempt to load the job info from the GP server
                self.job.get_info()
            except HTTPError:  # Handle HTTP errors contacting the server
                self.name = 'Error Loading Job'
                self.error = f'Error loading job #{self.job.job_number}'
                return

            # Add the job information to the widget
            self.name = f'Job #{self.job.job_number}'
            self.status = self.status_text()
            self.description = self.submitted_text()
            self.files = self.files_list()
            self.visualization = self.visualizer()

            # Send notification if completed
            self.handle_notification()

            # Update the menu items
            self.attach_terminate()
            self.extra_file_menu_items = {
                'Send to Code': {
                    'action': 'cell',
                    'code': f'import genepattern\n\nfile_{self.job.job_number} = genepattern.session.get("{self.job.server_data.url}").get_job({self.job.job_number}).get_file("{{{{file_name}}}}")\nfile_{self.job.job_number}'
                },
                'Send to Dataframe': {
                    'action': 'cell',
                    'kinds': ['gct', 'odf'],
                    'code': f'import genepattern\nfrom gp.data import GCT as gct, ODF as odf\n\nfile_{self.job.job_number} = {{{{type}}}}(genepattern.session.get("{self.job.server_data.url}").get_job({self.job.job_number}).get_file("{{{{file_name}}}}"))\nfile_{self.job.job_number}'
                },
                'Download to Workspace': {
                    'action': 'method',
                    'code': 'workspace_download',
                    'params': '{"file_name": "{{file_name}}"}'
                },
            }

            # Handle child jobs
            if len(self.appendix.children) == 0:
                self.appendix.children = [GPJobWidget(child) for child in self.job.get_child_jobs()]

            # Begin polling if pending or running
            self.poll_if_needed()
        else:
            # Display error message if no initialized GPJob object is provided
            self.name = 'Not Authenticated'
            self.error = 'You must be authenticated before the job can be displayed. After you authenticate it may take a few seconds for the information to appear.'

    def visualizer(self):
        if not self.initialized(): return  # Ensure the job has been set

        # Get the token, using the shim if necessary
        if hasattr(self.job.server_data, 'get_token'): token = self.job.server_data.get_token()
        else: token = get_token(self.job.server_data)

        # Handle server-relative URLs
        if 'launchUrl' in self.job.info:
            launch_url = self.job.info["launchUrl"]
            if launch_url[0] == '/': launch_url = launch_url[3:]
            launch_url = f'{self.job.server_data.url}{launch_url}#{token}'

            # Handle URLs with S3-redirects in the GET parameters
            parsed_url = urlparse(launch_url)
            params = parse_qs(parsed_url.query)
            for key in params.keys():
                if is_url(params[key][0]):
                    params[key][0] = redirect_url(params[key][0], token=token)
            parsed_url = parsed_url._replace(query=urlencode(params, doseq=True))

            return parsed_url.geturl()

        # Handle index.html or single HTML returns
        single_html = None
        for f in self.files:
            if f.endswith('/index.html'):
                return f
            elif f.endswith('.html') and single_html is None and not f.endswith('pipeline_execution_log.html'):
                single_html = f
            elif f.endswith('.html') and single_html is not None and not f.endswith('pipeline_execution_log.html'):
                single_html = False
        if single_html: return f'{single_html}#{token}'

        # Otherwise there is no visualizer
        return ''

    def poll_if_needed(self):
        """Begin a polling interval if the job is pending or running"""
        if self.status == 'Pending' or self.status == 'Running':
            timer = Timer(15.0, lambda: self.poll())
            timer.start()

    def submitted_text(self):
        """Return pretty job submission text"""
        if not self.initialized(): return  # Ensure the job has been set
        return f'Submitted by {self.job.info["userId"]} on {self.job.date_submitted}'

    def files_list(self):
        """Return the list of output and log files in the format the widget can handle"""
        if not self.initialized(): return  # Ensure the job has been set
        return [f['link']['href'] for f in (self.job.output_files + self.job.log_files)]

    def handle_notification(self):
        if self.status == 'Error':
            ToolManager.instance().send('notification', {'message': f'{self.name} has an error!', 'sender': 'GenePattern Notebook'})
        elif self.status == 'Completed':
            ToolManager.instance().send('notification', {'message': f'{self.name} is complete!', 'sender': 'GenePattern Notebook'})

    def status_text(self):
        """Return concise status text"""
        if not self.initialized(): return ''  # Ensure the job has been set
        if 'hasError' in self.job.info['status'] and self.job.info['status']['hasError']:
            return 'Error'
        elif 'completedInGp' in self.job.info['status'] and self.job.info['status']['completedInGp']:
            return 'Completed'
        elif 'isPending' in self.job.info['status'] and self.job.info['status']['isPending']:
            return 'Pending'
        else:
            return 'Running'

    def attach_detach(self):
        """Attach the menu option to detach the job widget from the analysis cell"""
        session_index = f'"{self.job.server_data.url}"' if self.job.server_data else 0
        self.extra_menu_items = {**self.extra_menu_items, **{'Detach Job': {
                'action': 'cell',
                'code': f'import gp\n\ngenepattern.display(gp.GPJob(genepattern.session.make({session_index}), {self.job.job_number}))'
            }}}

    def attach_sharing(self):
        if self.sharing_displayed: self.toggle_job_sharing()  # Display sharing if toggled on
        self.extra_menu_items = {**self.extra_menu_items, **{'Share Job': {
                'action': 'method',
                'code': 'toggle_job_sharing'
            }}}

    def attach_terminate(self):
        # Add the Terminate Job menu option, if necessaru
        if self.status == 'Pending' or self.status == 'Running' and 'Terminate Job' not in self.extra_menu_items:
            self.extra_menu_items = {**self.extra_menu_items, **{'Terminate Job': {
                    'action': 'method',
                    'code': 'terminate_job'
                }}}
        # Remove the Terminate Job menu option id no longer running or pending
        if self.status != 'Pending' and self.status != 'Running' and 'Terminate Job' in self.extra_menu_items:
            self.extra_menu_items = { key:value for key, value in self.extra_menu_items.items() if key != 'Terminate Job'}

    def attach_duplicate(self):
        session_index = f'"{self.job.server_data.url}"' if self.job.server_data else 0
        self.extra_menu_items = {**self.extra_menu_items, **{'Duplicate Analysis': {
            'action': 'cell',
            'code': f"genepattern.display(genepattern.reproduce_job(genepattern.session, {session_index}, {self.job.job_number}))"
        }}}

    def build_sharing_controls(self):
        """Create and return a VBox with the job sharing controls"""
        # Query job permissions, using the shim if necessary
        if hasattr(self.job, 'get_permissions'):
            perms = self.job.get_permissions()
        else:
            perms = get_permissions(self.job)
        group_widgets = []

        # Build the job sharing form by iterating over groups
        for g in perms['groups']:
            # Special case for * group
            group_label = g['id']
            if group_label == "*": group_label = 'All Users'

            d = Dropdown(description=group_label, options=['Private', 'Read', 'Read & Write'])
            if g['read'] and g['write']:
                d.value = 'Read & Write'
            elif g['read']:
                d.value = 'Read'
            group_widgets.append(d)

        # Sharing box title
        title = HTML('<hr/><h3>GenePattern Job Sharing</h3>')

        # Cancel / Close the sharing form functionality
        cancel_button = Button(description='Cancel')
        cancel_button.on_click(lambda b: self.toggle_job_sharing())

        # Save sharing permissions functionality
        def save_permissions(button):
            save_perms = []
            for g in group_widgets:
                # Special case for public jobs
                group_name = '*' if g.description == 'All Users' else g.description

                if g.value == 'Read & Write':
                    save_perms.append({'id': group_name, 'read': True, 'write': True})
                elif g.value == 'Read':
                    save_perms.append({'id': group_name, 'read': True, 'write': False})
                else:
                    save_perms.append({'id': group_name, 'read': False, 'write': False})
            # Save the permissions, using the shim if necessary
            if hasattr(self.job, 'set_permissions'):
                self.job.set_permissions(save_perms)
            else:
                set_permissions(self.job, save_perms)
            self.toggle_job_sharing()

        save_button = Button(description='Save', button_style='info')
        save_button.on_click(save_permissions)

        # Create the button HBox
        button_box = HBox(children=[cancel_button, save_button])

        # Create the job sharing box and attach to the job widget
        return VBox(children=[title] + group_widgets + [button_box])

    def toggle_job_sharing(self):
        """Toggle displaying the job sharing controls off and on"""
        # Handle None's
        if not self.initialized(): return  # Ignore this call if the job has not been properly initialized

        if self.sharing_displayed:
            # Add the old appendix children back to the widget if any exist, else simply remove the sharing box
            self.appendix.children = self.sharing_displayed if self.sharing_displayed is not True else []
            self.sharing_displayed = False
        else:
            # Create the job sharing box
            permissions_box = self.build_sharing_controls()
            # Save any child widgets in the appendix so that they're available when toggled back on
            self.sharing_displayed = self.appendix.children if self.appendix.children else True
            # Attach to the job widget
            self.appendix.children = [permissions_box]

    def terminate_job(self):
        if self.initialized():
            # Terminate using the shim if necessary
            self.job.terminate() if hasattr(self.job, 'terminate') else terminate_job(self.job)
            self.poll()

    def workspace_download(self, file_name="AG.chip.cvt.chip"):
        with open(file_name, 'wb') as f:
            with self.job.get_file(file_name).open() as g:
                f.write(g.read())

    def initialized(self):
        """Has the widget been initialized with session credentials"""
        return self.job and self.job.server_data and self.job.server_data.username

    def login_callback(self, data):
        """Callback for after a user authenticates"""
        if not self.job: return
        if not self.job.server_data or (not self.initialized() and data.url == self.job.server_data.url):
            self.job.server_data = data
            self.error = ''
            self.set_color()
            self.poll()



import gp
import IPython

from .jobwidget import GPJobWidget
from .authwidget import GPAuthWidget
from .taskwidget import GPTaskWidget, spec_to_kwargs


def display(content, **kwargs):
    """
    Display a widget, text or other media in a notebook without the need to import IPython at the top level.
    Also handles wrapping GenePattern Python Library content in widgets.
    :param content:
    :return:
    """
    if isinstance(content, gp.GPServer):
        IPython.display.display(GPAuthWidget(content))
    elif isinstance(content, gp.GPTask):
        if 'spec' in kwargs and isinstance(kwargs['spec'], gp.GPJobSpec): spec_to_kwargs(kwargs)
        IPython.display.display(GPTaskWidget(content, **kwargs))
    elif isinstance(content, gp.GPJob):
        IPython.display.display(GPJobWidget(content))
    else:
        IPython.display.display(content)


from setuptools import setup


__version__ = '23.11'


with open('README.md') as f:
    long_description = f.read()


setup(name='genepattern-notebook',
      packages=['genepattern'],
      version=__version__,
      description='GenePattern Notebook extension for JupyterLab',
      long_description=long_description,
      long_description_content_type="text/markdown",
      license='BSD',
      author='Thorin Tabor',
      author_email='tmtabor@cloud.ucsd.edu',
      url='https://github.com/genepattern/genepattern-notebook',
      download_url='https://github.com/genepattern/genepattern-notebook/archive/' + __version__ + '.tar.gz',
      keywords=['genepattern', 'genomics', 'bioinformatics', 'ipython', 'jupyter'],
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Intended Audience :: Science/Research',
          'Intended Audience :: Developers',
          'Topic :: Scientific/Engineering :: Bio-Informatics',
          'License :: OSI Approved :: BSD License',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: 3.9',
          'Framework :: Jupyter',
      ],
      install_requires=[
          'genepattern-python>=1.4.2',
          'nbtools>=23.4',
          'notebook>=5.0.0',
          'ipywidgets>=7.0.0',
          'pandas',
      ],
      data_files=[("share/jupyter/nbtools", ["nbtools/genepattern.json"])],
      normalize_version=False,
      )



import gp
from IPython.display import display
from urllib.error import HTTPError
from nbtools import UIBuilder, ToolManager, NBTool, EventManager
from .sessions import session
from .shim import login, system_message
from .taskwidget import TaskTool
from .utils import GENEPATTERN_LOGO, GENEPATTERN_SERVERS, server_name, session_color


REGISTER_EVENT = """
    const target = event.target;
    const widget = target.closest('.nbtools') || target;
    const server_input = widget.querySelector('input[type=text]');
    if (server_input) window.open(server_input.value + '/pages/registerUser.jsf');
    else console.warn('Cannot obtain GenePattern Server URL');"""


AUTO_LOGIN_CHECK = """
    const nameEQ = "GenePattern=";
    const ca = document.cookie.split(';');
    let cmatch = null;
    for (let i = 0; i < ca.length; i++) {
        let c = ca[i];
        while (c.charAt(0) === ' ') c = c.substring(1, c.length);
        if (c.indexOf(nameEQ) === 0) cmatch = c.substring(nameEQ.length, c.length)
    }
    if (cmatch === null) return;
    const parts = cmatch.split("|");
    if (parts.length <= 1) return;
    let [u, p] = [parts[0], atob(decodeURIComponent(parts[1]))];
    this.widget_dialog({
        'title': 'Log into GenePattern Server',
        'body': 'You have already authenticated with GenePattern Cloud. Would you like to automatically sign in now?',
        'button_label': 'Login',
        'callback': () => {
            this.model.get('form').get('children')[1].get('children')[1].set('value', u);
            this.model.get('form').get('children')[1].get('children')[1].save();
            this.model.get('form').get('children')[2].get('children')[1].set('value', p);
            this.model.get('form').get('children')[2].get('children')[1].save();
            this.element.querySelector("button.nbtools-run").click();
        }
    });"""


class GPAuthWidget(UIBuilder):
    """A widget for authenticating with a GenePattern server"""
    login_spec = {  # The display values for building the login UI
        'name': 'Login',
        'collapse': False,
        'display_header': False,
        'logo': GENEPATTERN_LOGO,
        'color': session_color(),
        'run_label': 'Log into GenePattern',
        'buttons': {
            'Register an Account': REGISTER_EVENT
        },
        'events': {
            'load': AUTO_LOGIN_CHECK
        },
        'parameters': {
            'server': {
                'name': 'GenePattern Server',
                'type': 'choice',
                'combo': True,
                'sendto': False,
                'default': GENEPATTERN_SERVERS['GenePattern Cloud'],
                'choices': GENEPATTERN_SERVERS
            },
            'username': {
                'name': 'Username',
                'sendto': False,
            },
            'password': {
                'name': 'Password',
                'type': 'password',
                'sendto': False,
            }
        }
    }

    def __init__(self, session=None, **kwargs):
        """Initialize the authentication widget"""

        # Assign the session object, lazily creating one if needed
        if session is None: self.session = gp.GPServer('', '', '')
        else: self.session = session

        # Set blank token
        self.token = None

        # Check to see if the provided session has valid credentials
        if self.has_credentials() and self.validate_credentials():
            self.prepare_session()

            # Display the widget with the system message and no form
            UIBuilder.__init__(self, lambda: None, name=self.session.username, subtitle=self.session.url,
                               display_header=False, display_footer=False, color=session_color(self.session.url),
                               collapsed=True, logo=GENEPATTERN_LOGO, **kwargs)

        # If not, prompt the user to login
        else:
            # Apply the display spec
            for key, value in self.login_spec.items(): kwargs[key] = value

            # Call the superclass constructor with the spec
            UIBuilder.__init__(self, self.login, **kwargs)

    def prepare_session(self):
        """Prepare a valid session by registering the session and modules"""
        self.register_session()     # Register the session with the SessionList
        self.register_modules()     # Register the modules with the ToolManager
        self.system_message()       # Display the system message
        self.trigger_login()        # Trigger login callbacks of job and task widgets

    def login(self, server, username, password):
        """Login to the GenePattern server"""
        # Assign login values to session
        self.session.url = server
        self.session.username = username
        self.session.password = password

        # Validate the provided credentials
        if self.validate_credentials():
            self.replace_widget()
            self.prepare_session()

    def has_credentials(self):
        """Test whether the session object is instantiated and whether a username and password have been provided"""
        if type(self.session) is not gp.GPServer: return False  # Test type
        if not self.session.url: return False                   # Test server url
        if not self.session.username: return False              # Test username
        if not self.session.password: return False              # Test password
        return True

    def validate_credentials(self):
        """Call gpserver.login() to validate the provided credentials"""
        try:
            # Check to see if gp library supports login, otherwise call login shim
            if hasattr(self.session, 'login'): self.token = self.session.login()
            else: self.token = login(self.session)
            gp.core.GP_JOB_TAG = 'GenePattern Notebook'  # Set tag for jobs
            return True
        except HTTPError:
            self.error = 'Invalid username or password. Please try again.'
            return False
        except BaseException as e:
            self.error = str(e)
            return False

    def replace_widget(self):
        """Replace the unauthenticated widget with the authenticated mode"""
        self.form.form.children[2].value = ''        # Blank password so it doesn't get serialized
        self.form.collapsed = True
        self.form.name = self.session.username
        self.form.subtitle = self.session.url
        self.form.display_header=False
        self.form.display_footer=False
        self.form.form.children = []

    def register_session(self):
        """Register the validated credentials with the SessionList"""
        self.session = session.register(self.session.url, self.session.username, self.session.password)

    def register_modules(self):
        """Get the list available modules and register widgets for them with the tool manager"""
        for task in self.session.get_task_list():
            tool = TaskTool(server_name(self.session.url), task)
            ToolManager.instance().register(tool)

    def system_message(self):
        if hasattr(self.session, 'system_message'): message = self.session.system_message()
        else: message = system_message(self.session)
        self.info = message

    def trigger_login(self):
        """Dispatch a login event after authentication"""
        EventManager.instance().dispatch("gp.login", self.session)


class AuthenticationTool(NBTool):
    """Tool wrapper for the authentication widget"""
    origin = '+'
    id = 'authentication'
    name = 'GenePattern Login'
    description = 'Log into a GenePattern server'
    load = lambda x: GPAuthWidget()


# Register the authentication widget
ToolManager.instance().register(AuthenticationTool())

